%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ctr_query/3
% ctr_query(+G,+S1,-S2)
ctr_query(G,S1,S2):-
        tr_logging(on),
        !,
        ( tr_logFileName(FileName)
                -> open(FileName,write,LogHandle)
                 ; open('tr.log',write,LogHandle)
        ),
        assert(tr_logHandle(LogHandle)),
        differential_state(S1,S1d),
        dispatch(G,S1d,S2).
ctr_query(G,S1,S2):-
	differential_state(S1,S1d),
        dispatch(G,S1d,S2).

% dispatch/3
% dispatch(+G,+S1,-S2)
dispatch(seq(G11a,G12),S1,S2):-
	!,
        tr_log(dispatch(seq(G11a,G12),S1,S2),Counter),
        debug_check_store_seen_state(S1),
        dispatch(G11a,S1,S3),
        debug_check_store_seen_state(S3),
        ctr_simplify(G12,G2),
        dispatch(G2,S3,S2),
        debug_check_store_seen_state(S2),
        tr_log_return(dispatch(seq(G11a,G12),S1,S2),Counter).
dispatch(ins(G1),S1,S2):-
        !,
        tr_log(dispatch(ins(G1),S1,S2),Counter),
        debug_check_store_seen_state(S1),
        ctr_atomic_query(ins(G1),S1,S2),
        debug_check_store_seen_state(S2),
        tr_log_return(dispatch(ins(G1),S1,S2),Counter).
dispatch(del(G1),S1,S2):-
        !,
        tr_log(dispatch(del(G1),S1,S2),Counter),
        debug_check_store_seen_state(S1),
        ctr_atomic_query(del(G1),S1,S2),
        debug_check_store_seen_state(S2),
        tr_log_return(dispatch(del(G1),S1,S2),Counter).
dispatch(G1,S1,S2):-
        tabled_predicate(G1),
        !,
        tr_log(dispatch(G1,S1,S2),Counter),
        absolute_state(S1,S1t),
        % puts the current state into tabled states,
        %       so we know exactly what states were tabled,
        %       used for tracing
        debug_check_store_seen_state(S1t),
        % puts the current state into the tabled states and returns the S1t
        %debug_check_store_input_tabled_shell_state(S1t),
        debug_check_store_tabled_shell_state(S1t),
        tabled_ctr_atomic_query(G1,S1t,S2t),
        differential_state(S2t,S2),
        debug_check_store_seen_state(S2t),
        debug_check_store_tabled_shell_state(S2t),
        tr_log_return(dispatch(G1,S1t,S2),Counter).
dispatch(G1,S1,S2):-
        tr_log(dispatch(G1,S1,S2),Counter),
        debug_check_store_seen_state(S1),
        ctr_atomic_query(G1,S1,S2),
        debug_check_store_seen_state(S2),
        tr_log_return(dispatch(G1,S1,S2),Counter).

:- table(tabled_ctr_atomic_query/3).
:- use_variant_tabling(tabled_ctr_atomic_query/3).
%:- use_subsumptive_tabling(tabled_ctr_atomic_query/3).
tabled_ctr_atomic_query(G1,S1,S2t):-
        tr_log(tabled_ctr_atomic_query(G1,S1t,S2t),Counter),
        tr_clause(G1,Body),
        differential_state(S1,S1t),
        dispatch(Body,S1t,S2),
        absolute_state(S2,S2t),
        tr_log_return(tabled_ctr_atomic_query(G1,S1t,S2t),Counter).

% ctr_atomic_query/1
% ctr_atomic_query(+G)
ctr_atomic_query(ins(Fact),S1,S2):-
        !,
        db_ins(Fact,S1,S2).
ctr_atomic_query(del(Fact),S1,S2):-
        !,
        db_del(Fact,S1,S2).
ctr_atomic_query(G1,S1,S1):-
        dynamic_predicate(G1),
        !,
        db_select(G1,S1).
ctr_atomic_query(G1,S1,S2):-
        tr_clause(G1,Body),
        dispatch(Body,S1,S2).
ctr_atomic_query(not(G1),S1,S1):-
        !,
        \+(dispatch(G1,S1,_S2)).
ctr_atomic_query(noop,S1,S1):-
        !.
ctr_atomic_query(big_query,S1,S1):-
        (for(_,1,100000),fail;true),
        !.
ctr_atomic_query(G1,S1,S1):-
        \+( tr_clause(G1,_Body) ),
        !,
        % Prolog predicate
        call(G1).

% ctr_simplify/2
% ctr_simplify(+InputFormula,-OutputFormula)
ctr_simplify(seq(state,G1),G2):-
        !,
        ctr_simplify(G1,G2).
ctr_simplify(conc(state,G1),G2):-
        !,
        ctr_simplify(G1,G2).
ctr_simplify(conc(G1,state),G2):-
        !,
        ctr_simplify(G1,G2).
ctr_simplify(G1,G2):-
        G1=G2.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
